"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .params import WhenTimePeriodParams
from .when import chck_time_period
from .when import when_time_period



__all__ = [
    'WhenTimePeriodParams',
    'chck_time_period',
    'when_time_period']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieDesire(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    desires = homie.desires
    desire = desires['jupiter']


    attrs = list(desire.__dict__)

    assert attrs == [
        '_HomieDesire__homie',
        '_HomieDesire__params',
        '_HomieDesire__name',
        '_HomieDesire__when']


    assert inrepr(
        'desire.HomieDesire object',
        desire)

    assert hash(desire) > 0

    assert instr(
        'desire.HomieDesire object',
        desire)


    assert desire.homie is homie

    assert desire.params is not None

    assert desire.name == 'jupiter'

    assert len(desire.whens) == 3

    assert len(desire.groups) == 2

    assert desire.state is None

    assert desire.level is None

    assert desire.scene == 'awaydark'

    assert desire.weight == 0

    assert desire.delay == 0

    assert desire.when is not None

    assert len(desire.when) == 3

    assert desire.delayed is False

    assert desire.outcomes == {
        'default': [False, True, True]}

    assert desire.outcome is False


    sample_path = (
        f'{SAMPLES}/desire'
        '/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=desire.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=desire.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_HomieDesire_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    desires = homie.desires
    delayed = desires['delayed']
    default = desires['default']


    assert homie.desired == {
        'jupiter_room': default,
        'jupiter_zone': default,
        'neptune_room': default,
        'neptune_zone': default}


    delayed.update_timer(
        f'-{delayed.delay}s')

    assert homie.desired == {
        'jupiter_room': delayed,
        'jupiter_zone': delayed,
        'neptune_room': delayed,
        'neptune_zone': delayed}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ...homie import Homie



def test_PhueDevice(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    devices = homie.phue_devices
    device = devices['jupiter_button']


    attrs = list(device.__dict__)

    assert attrs == [
        '_PhueDevice__homie',
        '_PhueDevice__params',
        '_PhueDevice__name',
        '_PhueDevice__source',
        '_PhueDevice__bridge']


    assert inrepr(
        'device.PhueDevice object',
        device)

    assert hash(device) > 0

    assert instr(
        'device.PhueDevice object',
        device)


    assert device.homie is homie

    assert device.params is not None

    assert device.name == 'jupiter_button'

    assert device.bridge is None

    assert device.source is None

    assert device.unique is None

    assert device.present is False

    assert device.connect is False

    assert device.changed is False

    assert device.sensors is None


    device.refresh_source()

    assert device.unique == (
        '8155e7b2-e89b-3b1d'
        '-80af-a937994d9a78')


    sample_path = (
        f'{SAMPLES}/device/source.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=device.source,
        replace=REPLACES)

    expect = prep_sample(
        content=device.source,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/device/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=device.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=device.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_PhueDevice_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    devices = homie.phue_devices
    device = devices['jupiter_button']


    device.params.phid = None
    device.params.label = 'Jupiter Button'

    device.refresh_source()

    assert device.bridge is not None
    assert device.source is not None
    assert device.present is True
    assert device.connect is True
    assert device.changed is not False
    assert len(device.changed) == 4
    assert device.sensors == {
        'button1': (
            'd31f216f-a9d3-3a53-'
            '913c-4060280dff75'),
        'button2': (
            '605aeecd-ad1e-34c2-'
            'b8ad-d24bc2dd0899'),
        'button3': (
            'a3b57110-bebf-3a39-'
            'bd21-4652f4a3d943'),
        'button4': (
            '4ab3f6fe-8773-36d9-'
            '9cdf-e937208c1de7')}

    device.params.phid = None
    device.params.label = 'noexst'

    device.refresh_source()

    assert device.bridge is None
    assert device.source is None
    assert device.present is False
    assert device.connect is False
    assert device.changed is False
    assert device.sensors is None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path



SAMPLES = (
    Path(__file__).parent
    .joinpath('samples'))



STATE_PHIDS = [

    ('e693d3a2-f4ef-36de'
     '-a583-e4fcc23f8f7c'),

    ('da4c6a04-c9e8-4915'
     '-9818-0f898029a1e8')]

STATE_PATHS = [

    ('https://192.168.1.10'
     '/clip/v2/resource'
     '/grouped_light'
     f'/{STATE_PHIDS[0]}'),

    ('https://192.168.2.10'
     '/clip/v2/resource'
     '/grouped_light'
     f'/{STATE_PHIDS[1]}')]



LEVEL_PHIDS = STATE_PHIDS

LEVEL_PATHS = [

    ('https://192.168.1.10'
     '/clip/v2/resource'
     '/grouped_light'
     f'/{LEVEL_PHIDS[0]}'),

    ('https://192.168.2.10'
     '/clip/v2/resource'
     '/grouped_light'
     f'/{LEVEL_PHIDS[1]}')]



SCENE_PHIDS = [

    ('5808a516-aab3-3ec3'
     '-8eee-4db5152b07b5'),

    ('9678ff8b-d452-49f3'
     '-861c-74e5c5b2ca7c')]

SCENE_PATHS = [

    ('https://192.168.1.10'
     '/clip/v2/resource/scene'
     f'/{SCENE_PHIDS[0]}'),

    ('https://192.168.2.10'
     '/clip/v2/resource/scene'
     f'/{SCENE_PHIDS[1]}')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.times import Times

from ..params import WhenTimePeriodParams



def test_WhenTimePeriodParams() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    anchor = Times('now')


    start = anchor.shift('-1h@s')
    stop = anchor.shift('+1h@s')

    params = WhenTimePeriodParams(
        start=start,
        stop=stop)

    assert params.start and params.stop

    assert params.start == start
    assert params.stop == stop


    start = anchor.shift('-1h@s')
    stop = anchor.shift('-2h@s')

    params = WhenTimePeriodParams(
        start=start,
        stop=stop)

    assert params.start and params.stop

    assert params.start == start
    assert params.stop >= start



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Timer
from encommon.times import Times
from encommon.types import setate
from encommon.types import striplower

from enconnect.philipshue import Bridge
from enconnect.philipshue import BridgeParams


if TYPE_CHECKING:
    from ..homie import Homie
    from ..homie.params import HOMIE_STATE



_FETCH = dict[str, Any]
_RAWDEV = dict[str, dict[str, Any]]



class PhueBridge:
    """
    Contain the relevant attributes about the related device.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: BridgeParams
    __bridge: Bridge

    __name: str

    __fetched: Optional[_FETCH]
    __timer: Timer
    __merged: Optional[_RAWDEV]


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='PhueBridge',
            name=name,
            status='initial')


        bridges = (
            homie.params
            .phue_bridges)

        assert bridges is not None

        params = bridges[name]


        self.__homie = homie
        self.__params = params
        self.__bridge = Bridge(params)
        self.__name = name
        self.__fetched = None
        self.__merged = None


        self.__timer = Timer(
            60, start='-60s')


        homie.log_d(
            base='PhueBridge',
            name=name,
            status='created')


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> BridgeParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def bridge(
        self,
    ) -> Bridge:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__bridge


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def connect(
        self,
    ) -> bool:
        """
        Return the boolean indicating connection is established.

        :returns: Boolean indicating connection is established.
        """

        return bool(self.__merged)


    def refresh_source(
        self,
    ) -> None:
        """
        Refresh the cached information for the remote upstream.
        """

        timer = self.__timer

        timer.update(
            f'-{int(timer.timer)}s')

        assert timer.ready(False)

        self.fetched
        self.merged

        assert not timer.ready()


    @property
    def fetched(
        self,
    ) -> _FETCH:
        """
        Collect the complete dump of all resources within bridge.

        :returns: Complete dump of all resources within bridge.
        """

        fetched = self.__fetched
        timer = self.__timer
        bridge = self.__bridge
        request = bridge.request

        ready = timer.ready(False)

        if fetched and not ready:
            return deepcopy(fetched)


        runtime = Times()

        try:

            response = request(
                'get', 'resource')

            response.raise_for_status()

            fetched = response.json()

            self.homie.log_i(
                base='PhueBridge',
                name=self.name,
                action='fetch',
                elapsed=runtime.since,
                status='success')

        except Exception as reason:  # NOCVR

            self.homie.log_e(
                base='PhueBridge',
                name=self.name,
                action='fetch',
                elapsed=runtime.since,
                status='failure',
                exc_info=reason)

            if fetched is None:
                raise

            return deepcopy(fetched)


        assert isinstance(fetched, dict)


        self.__fetched = fetched
        self.__merged = None

        timer.update('now')

        return deepcopy(fetched)


    @property
    def merged(
        self,
    ) -> _RAWDEV:
        """
        Process the response and perform common transformations.

        :returns: Compiled response from the upstream endpoint.
        """

        merged = self.__merged

        if merged is not None:
            return deepcopy(merged)

        fetched = self.fetched


        source = {
            x['id']: x for x in
            fetched['data']}

        origin = deepcopy(source)


        def _enhance() -> None:

            rtype = item['rtype']
            rid = item['rid']

            if 'taurus_' in rtype:
                return

            item['_source'] = (
                origin[rid])


        items1 = source.items()

        for key, value in items1:

            if 'services' not in value:
                continue

            items2 = value['services']

            for item in items2:
                _enhance()


        self.__merged = source

        return deepcopy(source)


    def get_source(
        self,
        phid: Optional[str] = None,
        label: Optional[str] = None,
        type: Optional[str] = None,
        grid: Optional[str] = None,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param phid: Used for filtering resources for matching.
        :param label: Used for filtering resources for matching.
        :param type: Used for filtering resources for matching.
        :param grid: Used for filtering resources for matching.
        :returns: Information for matching resource in upstream.
        """

        assert phid or label

        if phid is not None:
            return self.get_source_phid(
                phid, type, grid)

        if label is not None:
            return self.get_source_label(
                label, type, grid)

        return None  # NOCVR


    def get_source_phid(
        self,
        phid: str,
        type: Optional[str] = None,
        grid: Optional[str] = None,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param phid: Used for filtering resources for matching.
        :param type: Used for filtering resources for matching.
        :param grid: Used for filtering resources for matching.
        :returns: Information for matching resource in upstream.
        """

        found: list[_FETCH] = []

        items = self.merged.items()

        for _phid, fetch in items:

            _type = fetch['type']

            if type and _type != type:
                continue

            _grid: Optional[str] = (
                fetch.get('group', {})
                .get('rid'))

            if grid and _grid != grid:
                continue

            if _phid != phid:
                continue

            found.append(fetch)

        assert len(found) in [0, 1]

        return found[0] if found else None


    def get_source_label(
        self,
        label: str,
        type: Optional[str] = None,
        grid: Optional[str] = None,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param label: Used for filtering resources for matching.
        :param type: Used for filtering resources for matching.
        :param grid: Used for filtering resources for matching.
        :returns: Information for matching resource in upstream.
        """

        found: list[_FETCH] = []

        label = striplower(label)

        items = self.merged.items()

        for phid, fetch in items:

            _type = fetch['type']

            if type and _type != type:
                continue

            _grid: Optional[str] = (
                fetch.get('group', {})
                .get('rid'))

            if grid and _grid != grid:
                continue

            if 'metadata' not in fetch:
                continue

            metadata = fetch['metadata']

            if 'owner' in fetch:
                continue

            name = striplower(
                metadata['name'])

            if name != label:
                continue

            found.append(fetch)

        assert len(found) in [0, 1]

        return found[0] if found else None


    def state_get(
        self,
        light_phid: str,
    ) -> 'HOMIE_STATE':
        """
        Return the current state of the light within the bridge.

        :param light_phid: Unique identifier of light in bridge.
        :returns: Current state of the light within the bridge.
        """

        merged = self.merged

        light = merged[light_phid]

        state = light['on']['on']

        return 'on' if state else 'off'


    def state_set(
        self,
        light_phid: str,
        state: 'HOMIE_STATE',
    ) -> None:
        """
        Update the current state of the light within the bridge.

        :param light_phid: Unique identifier of light in bridge.
        :param state: Desired state for the lights within group.
        """

        self.homie.log_d(
            base='PhueBridge',
            action='state_set',
            light=light_phid,
            value=state,
            status='attempt')

        runtime = Times()


        path = (
            'resource/grouped_light'
            f'/{light_phid}')


        payload: dict[str, Any] = {}

        key = 'on/on'
        value = state == 'on'

        setate(payload, key, value)


        self.bridge.request(
            method='put',
            path=path,
            json=payload)


        self.homie.log_d(
            base='PhueBridge',
            action='state_set',
            light=light_phid,
            value=state,
            elapsed=runtime.since,
            status='success')


    def level_get(
        self,
        light_phid: str,
    ) -> int:
        """
        Return the current level of the light within the bridge.

        :param light_phid: Unique identifier of light in bridge.
        :returns: Current level of the light within the bridge.
        """

        merged = self.merged

        light = merged[light_phid]

        level = (
            light['dimming']
            ['brightness'])

        return int(level)


    def level_set(
        self,
        light_phid: str,
        level: int,
    ) -> None:
        """
        Update the current level of the light within the bridge.

        :param light_phid: Unique identifier of light in bridge.
        :param level: Desired level for the lights within group.
        """

        self.homie.log_d(
            base='PhueBridge',
            action='level_set',
            light=light_phid,
            value=level,
            status='attempt')

        runtime = Times()


        path = (
            'resource/grouped_light'
            f'/{light_phid}')


        payload: dict[str, Any] = {}

        key = 'dimming/brightness'
        value = int(level)

        setate(payload, key, value)


        self.bridge.request(
            method='put',
            path=path,
            json=payload)


        self.homie.log_d(
            base='PhueBridge',
            action='level_set',
            light=light_phid,
            value=level,
            elapsed=runtime.since,
            status='success')


    def scene_get(
        self,
        group_phid: str,
    ) -> Optional[str]:
        """
        Return the current scene of the group within the bridge.

        :param group_phid: Unique identifier of group in bridge.
        :returns: Current scene of the group within the bridge.
        """

        items = self.merged.items()

        for phid, fetch in items:

            if 'group' not in fetch:
                continue

            _group = fetch['group']
            _phid = _group['rid']

            if _phid != group_phid:
                continue

            status = fetch['status']
            active = status['active']

            if active != 'inactive':
                return phid

        return None


    def scene_set(
        self,
        scene_phid: str,
    ) -> None:
        """
        Update the current scene of the group within the bridge.

        :param scene_phid: Unique identifier of group in bridge.
        """

        self.homie.log_d(
            base='PhueBridge',
            action='scene_set',
            value=scene_phid,
            status='attempt')

        runtime = Times()


        path = (
            'resource/scene'
            f'/{scene_phid}')


        payload: dict[str, Any] = {}

        key = 'recall/action'
        value = 'active'

        setate(payload, key, value)


        self.bridge.request(
            method='put',
            path=path,
            json=payload)


        self.homie.log_d(
            base='PhueBridge',
            action='scene_set',
            value=scene_phid,
            elapsed=runtime.since,
            status='success')


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = deepcopy(
            self.params.model_dump())

        params['token'] = (
            '*' * len(params['token']))

        return {
            'name': self.name,
            'connect': self.connect,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.utils import save_text

from respx import MockRouter

from ..params import WhenUbiqClientParams
from ...conftest import config_factory
from ...conftest import homie_factory
from ...homie import HomieWhen
from ...homie import HomieWhenParams

if TYPE_CHECKING:
    from ...homie import Homie



def test_when_ubiq_client(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """


    ubiq_client = (
        WhenUbiqClientParams(
            routers=['jupiter'],
            clients=['laptop']))

    params = HomieWhenParams(
        ubiq_client=ubiq_client)

    when = HomieWhen(homie, params)

    assert when.outcome is True


    ubiq_client = (
        WhenUbiqClientParams(
            routers=['jupiter'],
            clients=['laptop'],
            since=10))

    params = HomieWhenParams(
        ubiq_client=ubiq_client)

    when = HomieWhen(homie, params)

    assert when.outcome is True



def test_when_ubiq_client_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    config = (
        'ubiq_clients:\n'
        '  noexst:\n'
        '    router: jupiter\n'
        '    ubid: noexst\n'
        '  doexst:\n'
        '    router: neptune\n'
        '    label: My Phone\n')


    Path.mkdir(
        tmp_path.joinpath('config'))

    save_text(
        f'{tmp_path}/config/test.yml',
        content=config)

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock)


    ubiq_client = (
        WhenUbiqClientParams(
            routers=['jupiter'],
            clients=['noexst']))

    params = HomieWhenParams(
        ubiq_client=ubiq_client)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    ubiq_client = (
        WhenUbiqClientParams(
            clients=['doexst']))

    params = HomieWhenParams(
        ubiq_client=ubiq_client)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    ubiq_client = (
        WhenUbiqClientParams(
            routers=['jupiter'],
            clients=['phone'],
            since=90000))

    params = HomieWhenParams(
        ubiq_client=ubiq_client)

    when = HomieWhen(homie, params)

    assert when.outcome is True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from httpx import Response

from respx import MockRouter

from . import LEVEL_PATHS
from . import LEVEL_PHIDS
from . import SAMPLES
from . import SCENE_PATHS
from . import SCENE_PHIDS
from . import STATE_PATHS
from . import STATE_PHIDS
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ...homie import Homie



def test_PhueBridge(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    bridges = homie.phue_bridges
    bridge = bridges['jupiter']


    attrs = list(bridge.__dict__)

    assert attrs == [
        '_PhueBridge__homie',
        '_PhueBridge__params',
        '_PhueBridge__bridge',
        '_PhueBridge__name',
        '_PhueBridge__fetched',
        '_PhueBridge__merged',
        '_PhueBridge__timer']


    assert inrepr(
        'bridge.PhueBridge object',
        bridge)

    assert hash(bridge) > 0

    assert instr(
        'bridge.PhueBridge object',
        bridge)


    assert bridge.homie is homie

    assert bridge.params is not None

    assert bridge.bridge is not None

    assert bridge.name == 'jupiter'

    assert bridge.connect is True


    sample_path = (
        f'{SAMPLES}/bridge/fetched.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=bridge.fetched,
        replace=REPLACES)

    expect = prep_sample(
        content=bridge.fetched,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/bridge/merged.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=bridge.merged,
        replace=REPLACES)

    expect = prep_sample(
        content=bridge.merged,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/bridge/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=bridge.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=bridge.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_PhueBridge_source(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    bridges = homie.phue_bridges
    bridge = bridges['jupiter']


    source = bridge.get_source(
        '8155e7b2-e89b-3b1d'
        '-80af-a937994d9a78')

    assert source is not None
    assert source['id'] == (
        '8155e7b2-e89b-3b1d'
        '-80af-a937994d9a78')



def test_PhueBridge_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    bridges = homie.phue_bridges
    bridge = bridges['jupiter']

    phid = (
        '8155e7b2-e89b-3b1d'
        '-80af-a937994d9a78')


    source = bridge.get_source(
        label='Jupiter Button')

    assert source is not None
    assert source['id'] == phid


    source = bridge.get_source(
        phid, type='device')

    assert source is not None
    assert source['id'] == phid


    group = homie.groups['jupiter_room']
    scene = homie.scenes['awake']


    source = bridge.get_source(
        label=scene.params.phue_label,
        type='scene',
        grid=group.phue_unique)

    assert source is not None
    assert source['id'] == (
        scene.phue_unique(group))


    source = bridge.get_source(
        phid=scene.phue_unique(group),
        type='scene',
        grid=group.phue_unique)

    assert source is not None
    assert source['id'] == (
        scene.phue_unique(group))



def test_PhueBridge_state(
    homie: 'Homie',
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    """

    groups = homie.groups
    bridges = homie.phue_bridges


    for path in STATE_PATHS:

        (respx_mock
         .put(path)
         .mock(Response(200)))


    bridge = bridges['jupiter']
    group = groups['jupiter_room']

    assert group.phue_light

    state = bridge.state_get(
        group.phue_light['id'])

    assert state is not None
    assert state == 'off'

    bridge.state_set(
        STATE_PHIDS[0], 'on')


    bridge = bridges['neptune']
    group = groups['neptune_room']

    assert group.phue_light

    state = bridge.state_get(
        group.phue_light['id'])

    assert state is not None
    assert state == 'off'

    bridge.state_set(
        STATE_PHIDS[1], 'on')



def test_PhueBridge_level(
    homie: 'Homie',
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    """

    groups = homie.groups
    bridges = homie.phue_bridges


    for path in LEVEL_PATHS:

        (respx_mock
         .put(path)
         .mock(Response(200)))


    bridge = bridges['jupiter']
    group = groups['jupiter_room']

    assert group.phue_light

    level = bridge.level_get(
        group.phue_light['id'])

    assert level is not None
    assert level == 0

    bridge.level_set(
        LEVEL_PHIDS[0], 100)


    bridge = bridges['neptune']
    group = groups['neptune_room']

    assert group.phue_light

    level = bridge.level_get(
        group.phue_light['id'])

    assert level is not None
    assert level == 0

    bridge.level_set(
        LEVEL_PHIDS[1], 100)




def test_PhueBridge_scene(
    homie: 'Homie',
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    """

    groups = homie.groups
    bridges = homie.phue_bridges


    for path in SCENE_PATHS:

        (respx_mock
         .put(path)
         .mock(Response(200)))


    bridge = bridges['jupiter']
    group = groups['jupiter_zone']

    assert group.phue_unique

    scene = bridge.scene_get(
        group.phue_unique)

    assert scene is not None
    assert scene[:8] == '35ac3411'

    bridge.scene_set(SCENE_PHIDS[0])


    bridge = bridges['neptune']
    group = groups['neptune_zone']

    assert group.phue_unique

    scene = bridge.scene_get(
        group.phue_unique)

    assert scene is not None
    assert scene[:8] == 'd8bc6c89'

    bridge.scene_set(SCENE_PHIDS[1])


    scene = bridge.scene_get('dne')

    assert scene is None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import TYPE_CHECKING

from encommon.types import getate

if TYPE_CHECKING:
    from ..homie import HomieWhat



def what_phue_motion(
    what: 'HomieWhat',
    event: dict[str, Any],
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param what: Primary class instance for the conditonal.
    :param event: Event which was yielded from the stream.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        what.params.phue_motion)

    assert params is not None

    _device = params.device
    _sensor = params.sensor


    devices = (
        what.homie.phue_devices)

    device = devices[_device]

    if not device.unique:
        device.refresh_source()

    unique = device.unique

    if unique is None:
        return False


    sensors = device.sensors

    assert sensors is not None

    sensor = sensors[_sensor]


    matched: list[bool] = []


    if 'data' not in event:
        return False

    for item in event['data']:

        if 'motion' not in item:
            continue

        _sensor = item['id']
        _unique = item['owner']['rid']

        if unique != _unique:
            continue

        if sensor != _sensor:
            continue  # NOCVR

        matched.append(True)


    return any(matched)



def chck_phue_motion(
    what: 'HomieWhat',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param what: Primary class instance for the conditonal.
    """

    params = (
        what.params.phue_motion)

    assert params is not None

    devices = (
        what.homie.phue_devices)


    _device = params.device

    assert _device in devices



def what_phue_button(
    what: 'HomieWhat',
    event: dict[str, Any],
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param what: Primary class instance for the conditonal.
    :param event: Event which was yielded from the stream.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        what.params.phue_button)

    assert params is not None

    _device = params.device
    _sensor = params.sensor
    _events = params.events


    devices = (
        what.homie.phue_devices)

    device = devices[_device]

    if not device.unique:
        device.refresh_source()

    unique = device.unique

    if unique is None:
        return False


    sensors = device.sensors

    assert sensors is not None

    sensor = sensors[_sensor]


    matched: list[bool] = []


    if 'data' not in event:
        return False

    for item in event['data']:

        if 'button' not in item:
            continue

        _sensor = item['id']
        _unique = item['owner']['rid']
        _event = getate(
            item, 'button/last_event')

        if unique != _unique:
            continue

        if sensor != _sensor:
            continue

        if _event not in _events:
            continue

        matched.append(True)


    return any(matched)



def chck_phue_button(
    what: 'HomieWhat',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param what: Primary class instance for the conditonal.
    """

    params = (
        what.params.phue_button)

    assert params is not None

    devices = (
        what.homie.phue_devices)


    _device = params.device

    assert _device in devices



def what_phue_contact(
    what: 'HomieWhat',
    event: dict[str, Any],
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param what: Primary class instance for the conditonal.
    :param event: Event which was yielded from the stream.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        what.params.phue_contact)

    assert params is not None

    _device = params.device
    _sensor = params.sensor


    devices = (
        what.homie.phue_devices)

    device = devices[_device]

    if not device.unique:
        device.refresh_source()

    unique = device.unique

    if unique is None:
        return False


    sensors = device.sensors

    assert sensors is not None

    sensor = sensors[_sensor]


    matched: list[bool] = []


    if 'data' not in event:
        return False

    for item in event['data']:

        if 'contact_report' not in item:
            continue

        _sensor = item['id']
        _unique = item['owner']['rid']

        if unique != _unique:
            continue

        if sensor != _sensor:
            continue  # NOCVR

        matched.append(True)


    return any(matched)



def chck_phue_contact(
    what: 'HomieWhat',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param what: Primary class instance for the conditonal.
    """

    params = (
        what.params.phue_contact)

    assert params is not None

    devices = (
        what.homie.phue_devices)


    _device = params.device

    assert _device in devices



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import TimerParams
from encommon.times.common import PARSABLE

from .what import HomieWhat
from .when import HomieWhen

if TYPE_CHECKING:
    from .homie import Homie
    from .params import HOMIE_STATE
    from .params import HomieActionParams
    from .params import HomieWhatParams
    from .params import HomieWhenParams



_OUTCOMES = dict[str, list[bool]]



class HomieAction:
    """
    Normalize the desired parameter across multiple products.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: 'HomieActionParams'

    __name: str

    __what: list[HomieWhat]
    __when: list[HomieWhen]


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='HomieAction',
            name=name,
            status='initial')


        actions = (
            homie.params.actions)

        assert actions is not None

        params = actions[name]


        self.__homie = homie
        self.__params = params
        self.__name = name


        what = self.what or []

        self.__what = [
            HomieWhat(homie, x)
            for x in what]


        when = self.when or []

        self.__when = [
            HomieWhen(homie, x)
            for x in when]


        self.__validate_params()

        homie.log_d(
            base='HomieAction',
            name=name,
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        groups = self.homie.groups
        _groups = self.params.groups

        for name in _groups:
            assert name in groups

        scenes = self.homie.scenes
        _scene = self.params.scene

        assert _scene in scenes


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieActionParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def whats(
        self,
    ) -> list[HomieWhat]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__what)


    @property
    def whens(
        self,
    ) -> list[HomieWhen]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__when)


    @property
    def groups(
        self,
    ) -> list[str]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.groups


    @property
    def state(
        self,
    ) -> Optional['HOMIE_STATE']:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.state


    @property
    def level(
        self,
    ) -> Optional[int]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.level


    @property
    def scene(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.scene


    @property
    def weight(
        self,
    ) -> int:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.weight


    @property
    def pause(
        self,
    ) -> int:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.pause


    @property
    def what(
        self,
    ) -> Optional[list['HomieWhatParams']]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.what


    @property
    def when(
        self,
    ) -> Optional[list['HomieWhenParams']]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.when


    @property
    def paused(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether action is paused.

        :returns: Boolean indicating whether action is paused.
        """

        homie = self.homie
        timers = homie.timers

        children = timers.children

        name = self.name
        pause = self.pause


        unique = f'action_{name}'

        if unique not in children:

            params = TimerParams(
                pause, f'-{pause}s')

            timers.create(
                unique, params)


        ready = timers.ready(
            unique, update=False)

        return not ready


    def update_timer(
        self,
        value: Optional[PARSABLE] = None,
    ) -> None:
        """
        Update the existing timer from mapping within the cache.

        :param value: Override the time updated for timer value.
        """

        homie = self.homie
        timers = homie.timers

        name = self.name
        unique = f'action_{name}'

        timers.update(
            unique, value or 'now')


    def delete_timer(
        self,
    ) -> None:
        """
        Update the existing timer from mapping within the cache.
        """

        homie = self.homie
        timers = homie.timers

        name = self.name
        unique = f'action_{name}'

        timers.delete(unique)


    @property
    def outcomes(
        self,
    ) -> Optional[_OUTCOMES]:
        """
        Return the dictionaries of boolean groups from conditons.

        .. note::
           This method is highly redundant to that in desire.py.

        :returns: Dictionaries of boolean groups from conditons.
        """

        matched: _OUTCOMES = {}


        for when in self.__when:

            outcome = when.outcome
            family = when.family

            if family not in matched:
                matched[family] = []

            (matched[family]
                .append(outcome))


        return matched or None


    @property
    def outcome(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether conditions matched.

        .. note::
           This method is highly redundant to that in desire.py.

        :returns: Boolean indicating whether conditions matched.
        """

        outcomes = self.outcomes

        if outcomes is None:
            return True

        matched: list[bool] = []


        items = outcomes.items()

        for family, when in items:

            matched.append(
                all(when)
                if family == 'default'
                else any(when))

        assert len(matched) >= 1


        return all(matched)


    def matches(
        self,
        event: dict[str, Any],
    ) -> list[bool]:
        """
        Return the boolean indicating whether conditions matched.

        :param event: Event which was yielded from the stream.
        :returns: Boolean indicating whether conditions matched.
        """

        matches: list[bool] = []


        for what in self.__what:

            match = what.match(event)

            matches.append(match)


        return matches


    def match(
        self,
        event: dict[str, Any],
    ) -> bool:
        """
        Return the boolean indicating whether conditions matched.

        :param event: Event which was yielded from the stream.
        :returns: Boolean indicating whether conditions matched.
        """

        matches = self.matches(event)

        return any(matches)


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        .. note::
           This method is highly redundant to that in desire.py.

        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        return {
            'name': self.name,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional

from encommon.config import Params as _Params

from enconnect.philipshue import (
    BridgeParams as PhueBridgeParams)
from enconnect.ubiquiti import (
    RouterParams as UbiqRouterParams)

from ..homie import HomieActionParams
from ..homie import HomieDesireParams
from ..homie import HomieGroupParams
from ..homie import HomieSceneParams
from ..philipshue import PhueDeviceParams
from ..ubiquiti import UbiqClientParams



_GROUPS = dict[str, HomieGroupParams]
_SCENES = dict[str, HomieSceneParams]
_DESIRES = dict[str, HomieDesireParams]
_ACTIONS = dict[str, HomieActionParams]

_PHUE_BRIDGES = dict[str, PhueBridgeParams]
_PHUE_DEVICES = dict[str, PhueDeviceParams]
_UBIQ_ROUTERS = dict[str, UbiqRouterParams]
_UBIQ_CLIENTS = dict[str, UbiqClientParams]



class Params(_Params, extra='forbid'):
    """
    Process and validate the core configuration parameters.

    .. note::
       Input parameters are not defined, check parent class.

    :param cache: Optional cache file but required if state
        between executions is required; when in production.
    :param dryrun: Determine we should not change anything.
    :param sargs: Additional arguments on the command line.
    :param groups: Dictionary of parameters for the groups.
    :param scenes: Dictionary of parameters for the scenes.
    :param desires: List of desired states and conditionals.
    :param actions: List of event actions and conditionals.
    :param phue_bridges: Paramters for the product bridges.
    :param phue_devices: Paramters for the product devices.
    :param ubiq_routers: Paramters for the product routers.
    :param ubiq_clients: Paramters for the product clients.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    cache: str = 'sqlite:///:memory:'

    dryrun: bool = False

    sargs: Optional[dict[str, Any]] = None

    groups: Optional[_GROUPS] = None
    scenes: Optional[_SCENES] = None
    desires: Optional[_DESIRES] = None
    actions: Optional[_ACTIONS] = None

    phue_bridges: Optional[_PHUE_BRIDGES] = None
    phue_devices: Optional[_PHUE_DEVICES] = None
    ubiq_routers: Optional[_UBIQ_ROUTERS] = None
    ubiq_clients: Optional[_UBIQ_CLIENTS] = None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr

from ..params import HomieWhenParams
from ..when import HomieWhen
from ...builtins import WhenTimePeriodParams

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieWhen(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    time_period = (
        WhenTimePeriodParams(
            start='-2d'))

    params = HomieWhenParams(
        time_period=time_period)

    when = HomieWhen(homie, params)


    attrs = list(when.__dict__)

    assert attrs == [
        '_HomieWhen__homie',
        '_HomieWhen__params']


    assert inrepr(
        'when.HomieWhen object',
        when)

    assert hash(when) > 0

    assert instr(
        'when.HomieWhen object',
        when)


    assert when.homie is homie

    assert when.params is not None

    assert when.negate is False

    assert when.family == 'default'

    assert when.outcome is True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Literal
from typing import Optional

from pydantic import BaseModel



class PhueDeviceParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param bridge: Name of bridge where the device belongs.
    :param phid: Unique identifier for device within bridge.
    :param label: Name of the device to find within bridge.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    bridge: str
    phid: Optional[str] = None

    label: Optional[str] = None


    def __init__(
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)

        phid = self.phid
        label = self.label

        assert phid or label



class WhatPhueMotionParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param device: Name of device in scope for the operation.
    :param sensor: Name of sensor in scope for the operation.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    device: str

    sensor: Literal['motion1']



class WhatPhueButtonParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param device: Name of device in scope for the operation.
    :param events: Name of event in scope for the operation.
    :param sensor: Name of sensor in scope for the operation.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    device: str

    sensor: Literal[
        'button1',
        'button2',
        'button3',
        'button4']

    events: list[Literal[
        'initial_press',
        'long_release',
        'short_release']]



class WhatPhueContactParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param device: Name of device in scope for the operation.
    :param sensor: Name of sensor in scope for the operation.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    device: str

    sensor: Literal['contact1']



class WhenPhueChangeParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param devices: List of devices in scope for conditional.
    :param sensors: List of sensors in scope for conditional.
    :param since: Minimum required time post change occurred.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    devices: list[str]

    sensors: list[Literal[
        'button1',
        'button2',
        'button3',
        'button4',
        'contact1',
        'motion1']]

    since: int = 0



class WhenPhueSceneParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param group: Group from wherein the scene is located.
    :param scenes: Name of the Homie scenes that will match.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    group: str
    scenes: list[str]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ...homie import Homie



def test_UbiqClient(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    clients = homie.ubiq_clients
    client = clients['jupiter_desktop']


    attrs = list(client.__dict__)

    assert attrs == [
        '_UbiqClient__homie',
        '_UbiqClient__params',
        '_UbiqClient__name',
        '_UbiqClient__sources',
        '_UbiqClient__routers']


    assert inrepr(
        'client.UbiqClient object',
        client)

    assert hash(client) > 0

    assert instr(
        'client.UbiqClient object',
        client)


    assert client.homie is homie

    assert client.params is not None

    assert client.name == 'jupiter_desktop'

    assert client.routers is None

    assert client.sources is None

    assert client.uniques is None

    assert client.present is False

    assert client.connect is False

    assert client.lastseen is None


    client.refresh_source()


    sample_path = (
        f'{SAMPLES}/client/sources.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=client.sources,
        replace=REPLACES)

    expect = prep_sample(
        content=client.sources,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/client/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=client.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=client.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_UbiqClient_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    clients = homie.ubiq_clients
    client = clients['jupiter_desktop']


    client.params.ubid = (
        '65d963e0e9285243cbbe6116')

    client.params.mac = None
    client.params.ip = None
    client.params.label = None

    client.refresh_source()

    assert client.routers is not None
    assert client.sources is not None
    assert client.present is not False
    assert len(client.present) == 1
    assert client.connect is not False
    assert len(client.connect) == 1


    client.params.ip = '192.168.1.100'

    client.params.ubid = None
    client.params.mac = None
    client.params.label = None

    client.refresh_source()

    assert client.routers is not None
    assert client.sources is not None
    # assert removed for mypy warning
    assert len(client.present) == 1
    # assert removed for mypy warning
    assert len(client.connect) == 1


    client.params.label = 'jupiter_desktop'

    client.params.ubid = None
    client.params.mac = None
    client.params.ip = None

    client.refresh_source()

    assert client.routers is not None
    assert client.sources is not None
    # assert removed for mypy warning
    assert len(client.present) == 1
    # assert removed for mypy warning
    assert len(client.connect) == 1


    client.params.label = 'noexst'

    client.params.ubid = None
    client.params.mac = None
    client.params.ip = None

    client.refresh_source()

    assert client.routers is None
    assert client.sources is None
    assert client.present is False
    assert client.connect is False


    client.params.router = 'noexst'

    client.refresh_source()

    assert client.routers is None
    assert client.sources is None
    assert client.present is False
    assert client.connect is False



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ... import PROJECT
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ..config import Config



def test_Config(
    tmp_path: Path,
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param config: Primary class instance for configuration.
    """

    replaces = REPLACES | {
        'pytemp': tmp_path,
        'PROJECT': PROJECT,
        'SAMPLES': SAMPLES}

    sample_path = (
        f'{SAMPLES}/config.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=config.config,
        replace=replaces)

    expect = prep_sample(
        content=config.config,
        replace=replaces)

    assert sample == expect



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional

from pydantic import BaseModel



class UbiqClientParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param router: Name of router where the client belongs.
    :param ubid: Unique identifier for client within router.
    :param mac: Identify the client by using the MAC address.
    :param ip: Identify the client by using the IP address.
    :param label: Name of the client to find within router.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    router: Optional[str] = None
    ubid: Optional[str] = None

    mac: Optional[str] = None
    ip: Optional[str] = None
    label: Optional[str] = None


    def __init__(
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)

        router = self.router
        ubid = self.ubid
        mac = self.mac
        ip = self.ip
        label = self.label

        if ubid is not None:
            assert router is not None

        if ubid is None:
            assert mac or ip or label



class WhenUbiqClientParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param routers: List of routers in scope for conditional.
    :param clients: List of clients in scope for conditional.
    :param since: Minimum required time post change occurred.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    routers: Optional[list[str]] = None
    clients: list[str]

    since: int = 0



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from os import environ
from pathlib import Path



PROJECT = Path(__file__).parent
WORKSPACE = PROJECT.parents[2]

VERSION = (
    PROJECT
    .joinpath('version.txt')
    .read_text(encoding='utf-8')
    .splitlines()[0].strip())

__version__ = VERSION



ENPYRWS = (
    environ.get('ENPYRWS') == '1')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from pathlib import Path

from encommon.times import Duration
from encommon.times import Times
from encommon.utils import read_text
from encommon.utils import save_text

from httpx import Response

from pytest import fixture

from respx import MockRouter

from .config import Config
from .homie import Homie
from .homie.test import (
    SAMPLES as CORE_SAMPLES)
from .philipshue.test import (
    SAMPLES as PHUE_SAMPLES)
from .ubiquiti.test import (
    SAMPLES as UBIQ_SAMPLES)



_ANCHOR = Times('-0s@s')

REPLACES = {


    '__STAMP_SMP__': (
        _ANCHOR.shift('-1h').simple),

    '__STAMP_SUB__': (
        _ANCHOR.shift('-1h').subsec),

    '__DURATION__': (
        str(Duration(
            Times(0).since,
            groups=2))),


    '__FSEEN_CUR__': (
        str(int(_ANCHOR.shift('-9d')))),

    '__LSEEN_CUR__': (
        str(int(_ANCHOR.shift('-1d')))),

    '__FSEEN_OLD__': (
        str(int(_ANCHOR.shift('-9d')))),

    '__LSEEN_OLD__': (
        str(int(_ANCHOR.shift('-1d')))),


    '__ASSOC_TIME__': (
        str(int(_ANCHOR.shift('-2d')))),


    '__DISCO_CUR__': (
        str(int(_ANCHOR.shift('-3h')))),

    '__DISCO_OLD__': (
        str(int(_ANCHOR.shift('-1d')))),


    '__START_DAY__': (
        _ANCHOR.shift('-0d@d').simple),

    '__STOP_DAY__': (
        _ANCHOR.shift('+1d@d').simple)}



def config_factory(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    save_text(
        f'{tmp_path}/config.yml',
        content=(
            'enconfig:\n'
            '  paths:\n'
            f'    - {tmp_path}\n'
            f'    - {CORE_SAMPLES}\n'
            f'    - {PHUE_SAMPLES}\n'
            f'    - {UBIQ_SAMPLES}\n'
            'enlogger:\n'
            '  stdo_level: info\n'))

    return Config(
        f'{tmp_path}/config.yml',
        sargs={'dryrun': False})



@fixture
def config(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    return config_factory(tmp_path)



def homie_factory(  # noqa: CFQ001
    config: 'Config',
    respx_mock: MockRouter,
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    .. note::
       Function has slowly evolved and grown over time, but
       should be refactored and simplified in the future.

    :param config: Primary class instance for configuration.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    homie = Homie(config)


    phue_paths = [
        ('https://192.168.1.10'
         '/clip/v2/resource'),
        ('https://192.168.2.10'
         '/clip/v2/resource')]

    phue_files = [
        f'{PHUE_SAMPLES}/jupiter.json',
        f'{PHUE_SAMPLES}/neptune.json']


    ubiq_paths = [

        ('https://192.168.1.1'
         '/api/auth/login'),
        ('https://192.168.1.1'
         '/proxy/network'
         '/api/s/default/rest/user'),
        ('https://192.168.1.1'
         '/proxy/network'
         '/api/s/default/stat/sta'),

        ('https://192.168.2.1'
         '/api/auth/login'),
        ('https://192.168.2.1'
         '/proxy/network'
         '/api/s/default/rest/user'),
        ('https://192.168.2.1'
         '/proxy/network'
         '/api/s/default/stat/sta')]

    ubiq_files = [

        (f'{UBIQ_SAMPLES}/jupiter'
         '/historic.json'),
        (f'{UBIQ_SAMPLES}/jupiter'
         '/realtime.json'),

        (f'{UBIQ_SAMPLES}/neptune'
         '/historic.json'),
        (f'{UBIQ_SAMPLES}/neptune'
         '/realtime.json')]


    def _replaces(
        path: str,
    ) -> str:

        content = read_text(path)

        items = REPLACES.items()

        for key, new in items:

            content = (
                content
                .replace(key, new))

        loaded = loads(content)

        assert isinstance(loaded, dict)

        return content


    dumped = _replaces(phue_files[0])

    (respx_mock
     .get(phue_paths[0])
     .mock(Response(
         status_code=200,
         content=dumped)))


    dumped = _replaces(phue_files[1])

    (respx_mock
     .get(phue_paths[1])
     .mock(Response(
         status_code=200,
         content=dumped)))


    phue_bridges = (
        homie.phue_bridges
        .values())

    for phue_bridge in phue_bridges:
        phue_bridge.refresh_source()


    (respx_mock
     .post(ubiq_paths[0])
     .mock(Response(
         status_code=200)))


    dumped = _replaces(ubiq_files[0])

    (respx_mock
     .get(ubiq_paths[1])
     .mock(side_effect=[
         Response(401),
         Response(
             status_code=200,
             content=dumped),
         Response(
             status_code=200,
             content=dumped)]))


    dumped = _replaces(ubiq_files[1])

    (respx_mock
     .get(ubiq_paths[2])
     .mock(Response(
         status_code=200,
         content=dumped)))


    (respx_mock
     .post(ubiq_paths[3])
     .mock(Response(
         status_code=200)))


    dumped = _replaces(ubiq_files[2])

    (respx_mock
     .get(ubiq_paths[4])
     .mock(side_effect=[
         Response(401),
         Response(
             status_code=200,
             content=dumped),
         Response(
             status_code=200,
             content=dumped)]))


    dumped = _replaces(ubiq_files[3])

    (respx_mock
     .get(ubiq_paths[5])
     .mock(Response(
         status_code=200,
         content=dumped)))


    ubiq_routers = (
        homie.ubiq_routers
        .values())

    for ubiq_router in ubiq_routers:
        ubiq_router.refresh_source()


    return homie



@fixture
def homie(
    config: 'Config',
    respx_mock: MockRouter,
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    return homie_factory(config, respx_mock)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ...homie import Homie



def test_UbiqRouter(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    routers = homie.ubiq_routers
    router = routers['jupiter']


    attrs = list(router.__dict__)

    assert attrs == [
        '_UbiqRouter__homie',
        '_UbiqRouter__params',
        '_UbiqRouter__router',
        '_UbiqRouter__name',
        '_UbiqRouter__fetched',
        '_UbiqRouter__merged',
        '_UbiqRouter__timer']


    assert inrepr(
        'router.UbiqRouter object',
        router)

    assert hash(router) > 0

    assert instr(
        'router.UbiqRouter object',
        router)


    assert router.homie is homie

    assert router.params is not None

    assert router.router is not None

    assert router.name == 'jupiter'

    assert router.connect is True


    sample_path = (
        f'{SAMPLES}/router/fetched.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=router.fetched,
        replace=REPLACES)

    expect = prep_sample(
        content=router.fetched,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/router/merged.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=router.merged,
        replace=REPLACES)

    expect = prep_sample(
        content=router.merged,
        replace=REPLACES)

    assert sample == expect


    sample_path = (
        f'{SAMPLES}/router/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=router.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=router.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_UbiqRouter_source(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    routers = homie.ubiq_routers
    router = routers['jupiter']


    source = router.get_source(
        '65d963e0e9285243cbbe6116')

    assert source is not None
    assert source['_id'] == (
        '65d963e0e9285243cbbe6116')



def test_UbiqRouter_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    routers = homie.ubiq_routers
    router = routers['jupiter']

    ubid = '65d963e0e9285243cbbe6116'


    source = router.get_source(
        '1a:01:68:00:11:00')

    assert source is not None
    assert source['_id'] == ubid


    source = router.get_source(
        '192.168.1.100')

    assert source is not None
    assert source['_id'] == ubid


    source = router.get_source(
        label='Jupiter Desktop')

    assert source is not None
    assert source['_id'] == ubid



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..params import WhenTimePeriodParams
from ...homie import HomieWhen
from ...homie import HomieWhenParams

if TYPE_CHECKING:
    from ...homie import Homie



def test_when_time_period(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """


    time_period = (
        WhenTimePeriodParams(
            start='+2d'))

    params = HomieWhenParams(
        time_period=time_period)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    time_period = (
        WhenTimePeriodParams(
            start='-2d'))

    params = HomieWhenParams(
        time_period=time_period)

    when = HomieWhen(homie, params)

    assert when.outcome is True


    time_period = (
        WhenTimePeriodParams(
            start='-2d',
            stop='-1d'))

    params = HomieWhenParams(
        time_period=time_period)

    when = HomieWhen(homie, params)

    assert when.outcome is False



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import TYPE_CHECKING

from ..philipshue import chck_phue_button
from ..philipshue import chck_phue_contact
from ..philipshue import chck_phue_motion
from ..philipshue import what_phue_button
from ..philipshue import what_phue_contact
from ..philipshue import what_phue_motion

if TYPE_CHECKING:
    from .homie import Homie
    from .params import HomieWhatParams



class HomieWhat:
    """
    Determine when condition is true and provide information.

    :param homie: Primary class instance for Homie Automate.
    :param params: Parameters for the conditional operations.
    """

    __homie: 'Homie'
    __params: 'HomieWhatParams'


    def __init__(
        self,
        homie: 'Homie',
        params: 'HomieWhatParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie
        self.__params = params

        self.__validate_params()


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        if self.params.phue_button:
            chck_phue_button(self)

        if self.params.phue_contact:
            chck_phue_contact(self)

        if self.params.phue_motion:
            chck_phue_motion(self)


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieWhatParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    def match(
        self,
        event: dict[str, Any],
    ) -> bool:
        """
        Return the boolean indicating whether condition matched.

        :param event: Event which was yielded from the stream.
        :returns: Boolean indicating whether condition matched.
        """

        matched: list[bool] = []

        if self.params.phue_button:
            matched.append(
                what_phue_button(self, event))

        if self.params.phue_motion:
            matched.append(
                what_phue_motion(self, event))

        if self.params.phue_contact:
            matched.append(
                what_phue_contact(self, event))

        outcome = bool(
            matched and any(matched))

        return outcome



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .config import Config
from .params import Params



__all__ = [
    'Config',
    'Params']
